from __future__ import print_function
import sys
import phat

if __name__=='__main__':
    test_data = (sys.argv[1:] and sys.argv[1]) or "../../examples/torus.bin"

    print("Reading test data %s in binary format ..." % test_data)

    boundary_matrix = phat.boundary_matrix()
    # This is broken for some reason
    if not boundary_matrix.load(test_data):
        print("Error: test data %s not found!" % test_data)
        sys.exit(1)

    error = False

    def compute_chunked(mat):
        return mat.compute_persistence_pairs(phat.reductions.chunk_reduction)

    print("Comparing representations using Chunk algorithm ...")
    print("Running Chunk - Sparse ...")
    sparse_boundary_matrix = phat.boundary_matrix(phat.representations.sparse_pivot_column, boundary_matrix)
    sparse_pairs = compute_chunked(sparse_boundary_matrix)

    print("Running Chunk - Heap ...")
    heap_boundary_matrix = phat.boundary_matrix(phat.representations.vector_heap, boundary_matrix)
    heap_pairs = compute_chunked(heap_boundary_matrix)

    print("Running Chunk - Full ...")
    full_boundary_matrix = phat.boundary_matrix(phat.representations.full_pivot_column, boundary_matrix)
    full_pairs = compute_chunked(full_boundary_matrix)

    print("Running Chunk - BitTree ...")
    bit_tree_boundary_matrix = phat.boundary_matrix(phat.representations.bit_tree_pivot_column, boundary_matrix)
    bit_tree_pairs = compute_chunked(bit_tree_boundary_matrix)

    print("Running Chunk - Vec_vec ...")
    vec_vec_boundary_matrix = phat.boundary_matrix(phat.representations.vector_vector, boundary_matrix)
    vec_vec_pairs = compute_chunked(vec_vec_boundary_matrix)

    print("Running Chunk - Vec_heap ...")
    vec_heap_boundary_matrix = phat.boundary_matrix(phat.representations.vector_heap, boundary_matrix)
    vec_heap_pairs = compute_chunked(vec_heap_boundary_matrix)

    print("Running Chunk - Vec_set ...")
    vec_set_boundary_matrix = phat.boundary_matrix(phat.representations.vector_set, boundary_matrix)
    vec_set_pairs = compute_chunked(vec_set_boundary_matrix)

    print("Running Chunk - Vec_list ...")
    vec_list_boundary_matrix = phat.boundary_matrix(phat.representations.vector_list, boundary_matrix)
    vec_list_pairs = compute_chunked(vec_list_boundary_matrix)

    if sparse_pairs != heap_pairs:
        print("Error: sparse and heap differ!", file=sys.stderr)
        error = True
    if heap_pairs != full_pairs:
        print("Error: heap and full differ!", file=sys.stderr)
        error = True
    if full_pairs != vec_vec_pairs:
        print("Error: full and vec_vec differ!", file=sys.stderr)
        error = True
    if vec_vec_pairs != vec_heap_pairs:
        print("Error: vec_vec and vec_heap differ!", file=sys.stderr)
        error = True
    if vec_heap_pairs != vec_set_pairs:
        print("Error: vec_heap and vec_set differ!", file=sys.stderr)
        error = True
    if vec_set_pairs != bit_tree_pairs:
        print("Error: vec_set and bit_tree differ!", file=sys.stderr)
        error = True
    if bit_tree_pairs != vec_list_pairs:
        print("Error: bit_tree and vec_list differ!", file=sys.stderr)
        error = True
    if vec_list_pairs != sparse_pairs:
        print("Error: vec_list and sparse differ!", file=sys.stderr)
        error = True
    if error:
        sys.exit(1)
    else:
        print("All results are identical (as they should be)")

    print("Comparing algorithms using BitTree representation ...")
    print("Running Twist - BitTree ...")

    def bit_tree_mat():
        return phat.boundary_matrix(phat.representations.bit_tree_pivot_column, boundary_matrix)

    reps = phat.representations
    reds = phat.reductions
    def pairs(mat, red):
        return mat.compute_persistence_pairs(red)

    twist_boundary_matrix = bit_tree_mat()
    twist_pairs = pairs(twist_boundary_matrix, reds.twist_reduction)

    print("Running Standard - BitTree ...")
    std_boundary_matrix = bit_tree_mat()
    std_pairs = pairs(std_boundary_matrix, reds.standard_reduction)

    print("Running Chunk - BitTree ...")
    chunk_boundary_matrix = bit_tree_mat()
    chunk_pairs = pairs(chunk_boundary_matrix, reds.chunk_reduction)

    print("Running Row - BitTree ...")
    row_boundary_matrix = bit_tree_mat()
    row_pairs = pairs(row_boundary_matrix, reds.row_reduction)

    print("Running Spectral sequence - BitTree ...")
    ss_boundary_matrix = bit_tree_mat()
    ss_pairs = pairs(ss_boundary_matrix, reds.spectral_sequence_reduction)

    if twist_pairs != std_pairs:
        print("Error: twist and standard differ!", file=sys.stderr)
        error = True
    if std_pairs != chunk_pairs:
        print("Error: standard and chunk differ!", file=sys.stderr)
        error = True
    if chunk_pairs != row_pairs:
        print("Error: chunk and row differ!", file=sys.stderr)
        error = True
    if row_pairs != ss_pairs:
        print("Error: row and spectral sequence differ!", file=sys.stderr)
        error = True
    if ss_pairs != twist_pairs:
        print("Error: spectral sequence and twist differ!", file=sys.stderr)
        error = True
    if error:
        sys.exit(1)
    else:
         print("All results are identical (as they should be)")

    print("Comparing primal and dual approach using Chunk - Full ...")

    primal_boundary_matrix = phat.boundary_matrix(reps.full_pivot_column, boundary_matrix)
    primal_pairs = primal_boundary_matrix.compute_persistence_pairs(reds.chunk_reduction)

    dual_boundary_matrix = phat.boundary_matrix(reps.full_pivot_column, boundary_matrix)
    dual_pairs = dual_boundary_matrix.compute_persistence_pairs_dualized()

    if primal_pairs != dual_pairs:
        print("Error: primal and dual differ!", file=sys.stderr)
        error = True


    if error:
        sys.exit(1)
    else:
        print("All results are identical (as they should be)")

    print("Testing vector<vector> interface ...")

    vector_vector_boundary_matrix = phat.boundary_matrix(phat.representations.bit_tree_pivot_column)

    vector_vector_boundary_matrix.columns = boundary_matrix.columns

    if vector_vector_boundary_matrix != boundary_matrix:
        print("Error: [load|save]_vector_vector bug", file=sys.stderr)
        error = True

    if error:
        sys.exit(1)
    else:
        print("Test passed!")





"""This file contains a simple example that demonstrates the usage of the library interface"""

if __name__ == "__main__":

    print("""
    we will build an ordered boundary matrix of this simplicial complex consisting of a single triangle: 
    
     3
     |\\
     | \\
     |  \\
     |   \\ 4
    5|    \\
     |     \\
     |  6   \\
     |       \\
     |________\\
     0    2    1

""")

    import phat

    boundary_matrix = phat.boundary_matrix()

    # or define a boundary matrix with the chosen internal representation
    # boundary_matrix = phat.boundary_matrix(representation = phat.representations.bit_tree_pivot_column)

    # set the respective columns -- (dimension, boundary) pairs
    boundary_matrix.columns = [ (0, []),
                                (0, []),
                                (1, [0,1]),
                                (0, []),
                                (1, [1,3]),
                                (1, [0,3]),
                                (2, [2,4,5])]

    # or equivalently, boundary_matrix = phat.boundary_matrix(representation = ..., columns = ...)
    # would combine the creation of the matrix and the assignment of the columns

    # print some information of the boundary matrix:
    print("\nThe boundary matrix has %d columns:" % len(boundary_matrix.columns))
    for col in boundary_matrix.columns:
        s = "Column %d represents a cell of dimension %d." % (col.index, col.dimension)
        if (col.boundary):
            s = s + " Its boundary consists of the cells " + " ".join([str(c) for c in col.boundary])
        print(s)
    print("Overall, the boundary matrix has %d entries." % len(boundary_matrix))

    pairs = boundary_matrix.compute_persistence_pairs()

    pairs.sort()

    print("\nThere are %d persistence pairs: " % len(pairs))
    for pair in pairs:
        print("Birth: %d, Death: %d" % pair)




"""Bindings for the Persistent Homology Algorithm Toolbox

PHAT is a tool for algebraic topology. It can be used via phat.py to compute
persistent (co)homology from boundary matrices, using various reduction
algorithms and column data representations.

Here is a simple example of usage.

We will build an ordered boundary matrix of this simplicial complex consisting of a single triangle::

     3
     |\\
     | \\
     |  \\
     |   \\ 4
    5|    \\
     |     \\
     |  6   \\
     |       \\
     |________\\
     0    2    1

Now the code::

    import phat

    # define a boundary matrix with the chosen internal representation
    boundary_matrix = phat.boundary_matrix(representation = phat.representations.vector_vector)

    # set the respective columns -- (dimension, boundary) pairs
    boundary_matrix.columns = [ (0, []),
                                (0, []),
                                (1, [0,1]),
                                (0, []),
                                (1, [1,3]),
                                (1, [0,3]),
                                (2, [2,4,5])]

    # or equivalently,
    # boundary_matrix = phat.boundary_matrix(representation = ...,
    #                                        columns = ...)
    # would combine the creation of the matrix and
    # the assignment of the columns

    # print some information of the boundary matrix:
    print()
    print("The boundary matrix has %d columns:" % len(boundary_matrix.columns))
    for col in boundary_matrix.columns:
        s = "Column %d represents a cell of dimension %d." % (col.index, col.dimension)
        if (col.boundary):
            s = s + " Its boundary consists of the cells " + " ".join([str(c) for c in col.boundary])
        print(s)
    print("Overall, the boundary matrix has %d entries." % len(boundary_matrix))

    pairs = boundary_matrix.compute_persistence_pairs()

    pairs.sort()

    print()
    print("There are %d persistence pairs: " % len(pairs))
    for pair in pairs:
        print("Birth: %d, Death: %d" % pair)


Please see https://bitbucket.org/phat-code/phat/python for more information.
"""

import _phat
import enum

from _phat import persistence_pairs

# The public API for the module

__all__ = ['boundary_matrix',
           'persistence_pairs',
           'representations',
           'reductions']


class representations(enum.Enum):
    """Available representations for internal storage of columns in
    a `boundary_matrix`
    """
    bit_tree_pivot_column = 1
    sparse_pivot_column = 2
    full_pivot_column = 3
    vector_vector = 4
    vector_heap = 5
    vector_set = 6
    vector_list = 7


class reductions(enum.Enum):
    """Available reduction algorithms"""
    twist_reduction = 1
    chunk_reduction = 2
    standard_reduction = 3
    row_reduction = 4
    spectral_sequence_reduction = 5


class column(object):
    """A view on one column of data in a boundary matrix"""

    def __init__(self, matrix, index):
        """INTERNAL. Columns are created automatically by boundary matrices.
        There is no need to construct them directly"""
        self._matrix = matrix
        self._index = index

    @property
    def index(self):
        """The 0-based index of this column in its boundary matrix"""
        return self._index

    @property
    def dimension(self):
        """The dimension of the column (0 = point, 1 = line, 2 = triangle, etc.)"""
        return self._matrix._matrix.get_dim(self._index)

    @dimension.setter
    def dimension(self, value):
        self._matrix._matrix.set_dim(self._index, value)

    @property
    def boundary(self):
        """The boundary values in this column, i.e. the other columns that this column is bounded by"""
        return self._matrix._matrix.get_col(self._index)

    @boundary.setter
    def boundary(self, values):
        self._matrix._matrix.set_col(self._index, values)

    def __str__(self):
        return "(%d, %s)" % (self.dimension, self.boundary)


class boundary_matrix(object):
    """Boundary matrices that store the shape information of a cell complex.
    """

    def __init__(self, representation=representations.bit_tree_pivot_column, source=None, columns=None):
        """
        The boundary matrix will use the specified implementation for storing its
        column data. If the `source` parameter is specified, it will be assumed to
        be another boundary matrix, whose data should be copied into the new
        matrix.

        Parameters
        ----------

        representation : phat.representation, optional
            The type of column storage to use in the requested boundary matrix.
        source : phat.boundary_matrix, optional
            If provided, creates the requested matrix as a copy of the data and dimensions
            in `source`.
        columns : column list, or list of (dimension, boundary) tuples, optional
            If provided, loads these columns into the new boundary matrix. Note that
            columns will be loaded in the order given, not according to their ``index`` properties.

        Returns
        -------

        matrix : boundary_matrix
        """
        self._representation = representation
        if source:
            self._matrix = _convert(source, representation)
        else:
            self._matrix = self.__matrix_for_representation(representation)()
            if columns:
                self.columns = columns

    @property
    def columns(self):
        """A collection of column objects"""
        return [column(self, i) for i in range(self._matrix.get_num_cols())]

    @columns.setter
    def columns(self, columns):
        for col in columns:
            if not (isinstance(col, column) or isinstance(col, tuple)):
                raise TypeError("All columns must be column objects, or (dimension, values) tuples")
        if len(columns) != len(self.dimensions):
            self._matrix.set_dims([0] * len(columns))
        for i, col in enumerate(columns):
            if isinstance(col, column):
                self._matrix.set_dim(i, col.dimension)
                self._matrix.set_col(i, col.boundary)
            else:
                dimension, values = col
                self._matrix.set_dim(i, dimension)
                self._matrix.set_col(i, sorted(values))

    @property
    def dimensions(self):
        """A collection of dimensions, equivalent to [c.dimension for c in self.columns]"""
        return [self._matrix.get_dim(i) for i in range(self._matrix.get_num_cols())]

    @dimensions.setter
    def dimensions(self, dimensions):
        self._matrix.set_dims(dimensions)

    def __matrix_for_representation(self, representation):
        short_name = _short_name(representation.name)
        return getattr(_phat, "boundary_matrix_" + short_name)

    def __eq__(self, other):
        return self._matrix == other._matrix

    # Note Python 2.7 needs BOTH __eq__ and __ne__ otherwise you get things that
    # are both equal and not equal
    def __ne__(self, other):
        return self._matrix != other._matrix

    def __len__(self):
        return self._matrix.get_num_entries()

    # Pickle support
    def __getstate__(self):
        (dimensions, columns) = self._matrix.get_vector_vector()
        return (self._representation, dimensions, columns)

    # Pickle support
    def __setstate__(self, state):
        presentation, dimensions, columns = state
        self._representation = representation
        self._matrix = self.__matrix_for_representation(representation)
        self._matrix.set_vector_vector(dimensions, columns)

    def load(self, file_name, mode='b'):
        """Load this boundary matrix from a file

        Parameters
        ----------

        file_name : string
            The file name to load

        mode : string, optional (defaults to 'b')
            The mode ('b' for binary, 't' for text) to use for working with the file

        Returns
        -------

        success : bool

        """
        if mode == 'b':
            return self._matrix.load_binary(file_name)
        elif mode == 't':
            return self._matrix.load_ascii(file_name)
        else:
            raise ValueError("Only 'b' - binary and 't' - text modes are supported")

    def save(self, file_name, mode='b'):
        """Save this boundary matrix to a file

        Parameters
        ----------

        file_name : string
            The file name to load

        mode : string, optional (defaults to 'b')
            The mode ('b' for binary, 't' for text) to use for working with the file

        Returns
        -------

        success : bool

        """
        if mode == 'b':
            return self._matrix.save_binary(file_name)
        elif mode == 't':
            return self._matrix.save_ascii(file_name)
        else:
            raise ValueError("Only 'b' - binary and 't' - text modes are supported")

    def compute_persistence_pairs(self,
                                  reduction=reductions.twist_reduction):
        """Computes persistence pairs (birth, death) for the given boundary matrix."""
        representation_short_name = _short_name(self._representation.name)
        algo_name = reduction.name
        algo_short_name = _short_name(algo_name)
        # Look up an implementation that matches the requested characteristics
        # in the _phat module
        function = getattr(_phat, "compute_persistence_pairs_" + representation_short_name + "_" + algo_short_name)
        return function(self._matrix)

    def compute_persistence_pairs_dualized(self,
                                           reduction=reductions.twist_reduction):
        """Computes persistence pairs (birth, death) from the dualized form of the given boundary matrix."""
        representation_short_name = _short_name(self._representation.name)
        algo_name = reduction.name
        algo_short_name = _short_name(algo_name)
        # Look up an implementation that matches the requested characteristics
        # in the _phat module
        function = getattr(_phat,
                           "compute_persistence_pairs_dualized_" + representation_short_name + "_" + algo_short_name)
        return function(self._matrix)

    def convert(self, representation):
        """Copy this matrix to another with a different representation"""
        return boundary_matrix(representation, self)


def _short_name(name):
    """An internal API that takes leading characters from words
    For instance, 'bit_tree_pivot_column' becomes 'btpc'
    """
    return "".join([n[0] for n in name.split("_")])


def _convert(source, to_representation):
    """Internal - function to convert from one `boundary_matrix` implementation to another"""
    class_name = source._representation.name
    source_rep_short_name = _short_name(class_name)
    to_rep_short_name = _short_name(to_representation.name)
    function = getattr(_phat, "convert_%s_to_%s" % (source_rep_short_name, to_rep_short_name))
    return function(source._matrix)



from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
import sys
import os.path
from io import open

if sys.version_info < (2, 7, 12):
    print("Sorry, PHAT requires Python 2.7.12 or later")
    sys.exit(1)


ext_modules = [
    Extension(
        '_phat',
        ['python/_phat.cpp'],
        include_dirs=['include'],
        language='c++',
    ),
]

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(here, 'python', 'README.rst'), encoding = 'utf8') as f:
    long_description = f.read()

class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': ['-std=c++11'],
    }

    if sys.platform == 'darwin':
        c_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        import pybind11
        for ext in self.extensions:
            ext.extra_compile_args = opts
            ext.include_dirs.append(pybind11.get_include())
            ext.include_dirs.append(pybind11.get_include(user=True))
        build_ext.build_extensions(self)

requires = ['pybind11']

if sys.version_info < (3,4,0):
    requires.append('enum34')

setup(
    name='phat',
    version='1.6.0.post2',
    author='Bryn Keller',
    author_email='bryn.keller@intel.com',
    url='https://bitbucket.org/phat-code/phat',
    description='Python bindings for PHAT, the C++ based Persistent Homology Algorithm Toolbox',
    license = 'LGPL',
    keywords='algebraic-topology PHAT distributed topology persistent-homology',
    long_description=long_description,
    ext_modules=ext_modules,
    requires=requires,
    install_requires=requires,
    cmdclass={'build_ext': BuildExt},
    package_dir={'':'python'},
    py_modules = ['phat'],
    # packages = find_packages(exclude = ['doc', 'test'])
 )





